require "state"
require "permutation"
require "orientation"
require "combination"
require "idastar"


-- indexing

local function state_to_indices(state)
    local result = {}

    -- corners orientation
    result.oCorners = Orientation.to_index(state.oCorners, 3)

    -- edges orientation
    result.oEdges = Orientation.to_index(state.oEdges, 2)

    -- e edges combination
    local e_edges = {
        true,  true,  true,  true,
        false, false, false, false,
        false, false, false, false,
    }

    local cEdges = {}
    for i = 1, 12 do
        cEdges[i] = e_edges[state.pEdges[i]]
    end

    result.cEdges = Combination.to_index(cEdges, 4)

    return result
end

local function indices_to_state(indices)
    -- corners orientation
    local oCornersIndex = indices.oCorners or 0
    local oCorners = Orientation.from_index(oCornersIndex, 3, 8)

    -- edges orientation
    local oEdgesIndex = indices.oEdges or 0
    local oEdges = Orientation.from_index(oEdgesIndex, 2, 12)

    -- e edges combination
    local cEdgesIndex = indices.cEdges or 0
    local cEdges = Combination.from_index(cEdgesIndex, 4, 12)

    local next_e = { 1, 2, 3, 4 }
    local next_not_e = { 5, 6, 7, 8, 9, 10, 11, 12 }

    local pEdges = {}
    for i = 1, 12 do
        if cEdges[i] then
            pEdges[i] = table.remove(next_e)
        else
            pEdges[i] = table.remove(next_not_e)
        end
    end

    return State:new(
        { 1, 2, 3, 4, 5, 6, 7, 8 },
        oCorners,
        pEdges,
        oEdges)
end


-- moves

local moves = {
    ["U"]  = State.moves["U"],
    ["U2"] = State.moves["U2"],
    ["U'"] = State.moves["U'"],
    ["D"]  = State.moves["D"],
    ["D2"] = State.moves["D2"],
    ["D'"] = State.moves["D'"],
    ["L"]  = State.moves["L"],
    ["L2"] = State.moves["L2"],
    ["L'"] = State.moves["L'"],
    ["R"]  = State.moves["R"],
    ["R2"] = State.moves["R2"],
    ["R'"] = State.moves["R'"],
    ["F"]  = State.moves["F"],
    ["F2"] = State.moves["F2"],
    ["F'"] = State.moves["F'"],
    ["B"]  = State.moves["B"],
    ["B2"] = State.moves["B2"],
    ["B'"] = State.moves["B'"],
}


-- move tables

local N_CORNERS_ORIENTATIONS = Orientation.number(3, 8)

local oCornersMove = {}
for i = 0, N_CORNERS_ORIENTATIONS - 1 do
    local state = indices_to_state({ oCorners = i })

    oCornersMove[i] = {}
    for move_name, move in pairs(moves) do
        oCornersMove[i][move_name] =
            state_to_indices(state:mul(move)).oCorners
    end
end


local N_EDGES_ORIENTATIONS = Orientation.number(2, 12)

local oEdgesMove = {}
for i = 0, N_EDGES_ORIENTATIONS - 1 do
    local state = indices_to_state({ oEdges = i })

    oEdgesMove[i] = {}
    for move_name, move in pairs(moves) do
        oEdgesMove[i][move_name] =
            state_to_indices(state:mul(move)).oEdges
    end
end


local N_EDGES_COMBINATIONS = Combination.number(4, 12)

local cEdgesMove = {}
for i = 0, N_EDGES_COMBINATIONS - 1 do
    local state = indices_to_state({ cEdges = i })

    cEdgesMove[i] = {}
    for move_name, move in pairs(moves) do
        cEdgesMove[i][move_name] =
            state_to_indices(state:mul(move)).cEdges
    end
end


-- distance tables

local goal = state_to_indices(State.id)

local oCornersDistance = {}
for i = 0, N_CORNERS_ORIENTATIONS - 1 do
    oCornersDistance[i] = {}
end

oCornersDistance[goal.oCorners][goal.cEdges] = 0

local distance = 0
repeat
    local n_visited = 0

    for i = 0, N_CORNERS_ORIENTATIONS - 1 do
        for j = 0, N_EDGES_COMBINATIONS - 1 do
            if oCornersDistance[i][j] == distance then
                for move_name, _ in pairs(moves) do
                    local next_o_corners = oCornersMove[i][move_name]
                    local next_c_edges = cEdgesMove[j][move_name]
                    if not oCornersDistance[next_o_corners][next_c_edges] then
                        oCornersDistance[next_o_corners][next_c_edges] = distance + 1
                        n_visited = n_visited + 1
                    end
                end
            end
        end
    end

    distance = distance + 1
until n_visited == 0


local oEdgesDistance = {}
for i = 0, N_EDGES_ORIENTATIONS - 1 do
    oEdgesDistance[i] = {}
end

oEdgesDistance[goal.oEdges][goal.cEdges] = 0

local distance = 0
repeat
    local n_visited = 0

    for i = 0, N_EDGES_ORIENTATIONS - 1 do
        for j = 0, N_EDGES_COMBINATIONS - 1 do
            if oEdgesDistance[i][j] == distance then
                for move_name, _ in pairs(moves) do
                    local next_o_edges = oEdgesMove[i][move_name]
                    local next_c_edges = cEdgesMove[j][move_name]
                    if not oEdgesDistance[next_o_edges][next_c_edges] then
                        oEdgesDistance[next_o_edges][next_c_edges] = distance + 1
                        n_visited = n_visited + 1
                    end
                end
            end
        end
    end

    distance = distance + 1
until n_visited == 0


-- search

Phase1 = IDAStar:new()

function Phase1:solved(state)
    return state.oCorners == goal.oCorners and
           state.oEdges == goal.oEdges and
           state.cEdges == goal.cEdges
end

function Phase1:prune(state, depth)
    return oCornersDistance[state.oCorners][state.cEdges] > depth or
           oEdgesDistance[state.oEdges][state.cEdges] > depth
end

local axis = {
    ["U"] = "U", ["U2"] = "U", ["U'"] = "U",
    ["D"] = "D", ["D2"] = "D", ["D'"] = "D",
    ["L"] = "L", ["L2"] = "L", ["L'"] = "L",
    ["R"] = "R", ["R2"] = "R", ["R'"] = "R",
    ["F"] = "F", ["F2"] = "F", ["F'"] = "F",
    ["B"] = "B", ["B2"] = "B", ["B'"] = "B",
}

function Phase1:next_states(state, path)
    return coroutine.wrap(function()
        local last = path[table.maxn(path)]

        for move_name, move in pairs(moves) do
            if not last or axis[last] ~= axis[move_name] then
                local next_state = {
                    oCorners = oCornersMove[state.oCorners][move_name],
                    oEdges = oEdgesMove[state.oEdges][move_name],
                    cEdges = cEdgesMove[state.cEdges][move_name],
                }

                coroutine.yield(move_name, next_state)
            end
        end
    end)
end

function Phase1:preprocess_state(state)
    return state_to_indices(state)
end
