require "state"
require "permutation"
require "orientation"
require "combination"


-- indexing

local function state_to_indices(state)
	local result = {}

	-- corners permutation
	result.pCorners = Permutation.to_index(state.pCorners)

	-- u d edges permutation
	local ud_edges = {
		false, false, false, false,
		true, true, true, true,
		true, true, true, true,
	}

	local ud_mapping = {
		0, 0, 0, 0,
		1, 2, 3, 4,
		5, 6, 7, 8,
	}

	local pUDEdges = {}
	for i = 1, 12 do
		if ud_edges[i] then
			table.insert(pUDEdges, ud_mapping[state.pEdges[i]])
		end
	end

	result.pUDEdges = Permutation.to_index(pUDEdges)

	-- e edges permutation
	local e_edges = {
        true,  true,  true,  true,
		false, false, false, false,
		false, false, false, false,
	}

	local e_mapping = {
		1, 2, 3, 4,
		0, 0, 0, 0,
		0, 0, 0, 0,
	}

	local pEEdges = {}
	for i = 1, 12 do
		if e_edges[i] then
			table.insert(pEEdges, e_mapping[state.pEdges[i]])
		end
	end

	result.pEEdges = Permutation.to_index(pEEdges)

	return result
end

local function indices_to_state(indices)
	-- corners permutation
	local pCornersIndex = indices.pCorners or 0
	local pCorners = Permutation.from_index(pCornersIndex, 8)

	-- edges permutation
	local pUDEdgesIndex = indices.pUDEdges or 0
	local pUDEdges = Permutation.from_index(pUDEdgesIndex, 8)

	local pEEdgesIndex = indices.pEEdges or 0
	local pEEdges = Permutation.from_index(pEEdgesIndex, 4)

	local ud_edges = {
		false, false, false, false,
		true,  true,  true,  true,
		true,  true,  true,  true,
	}

	local ud_edges_mapping = {
		5,  6,  7,  8,
		9, 10, 11, 12,
	}
	local ud_next = 1

	local e_edges = {
        true,  true,  true,  true,
		false, false, false, false,
		false, false, false, false,
	}

	local e_edges_mapping = {
		1,  2,  3,  4,
	}
	local e_next = 1

	local pEdges = {}

	-- u d edges
	for i = 1, 12 do
		if ud_edges[i] then
			pEdges[i] = ud_edges_mapping[pUDEdges[ud_next]]
			ud_next = ud_next + 1
		else
			pEdges[i] = e_edges_mapping[pEEdges[e_next]]
			e_next = e_next + 1
		end
	end

	return State:new(
		pCorners,
		{ 0, 0, 0, 0, 0, 0, 0, 0 },
		pEdges,
		{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 })
end


-- moves

local moves = {
	["U"]  = State.moves["U"],
	["U2"] = State.moves["U2"],
	["U'"] = State.moves["U'"],
	["D"]  = State.moves["D"],
	["D2"] = State.moves["D2"],
	["D'"] = State.moves["D'"],
	["L2"] = State.moves["L2"],
	["R2"] = State.moves["R2"],
	["F2"] = State.moves["F2"],
	["B2"] = State.moves["B2"],
}


-- move tables

local N_CORNERS_PERMUTATIONS = Permutation.number(8)

local pCornersMove = {}
for i = 0, N_CORNERS_PERMUTATIONS - 1 do
	local state = indices_to_state({ pCorners = i })

	pCornersMove[i] = {}
	for move_name, move in pairs(moves) do
		pCornersMove[i][move_name] =
			state_to_indices(state:mul(move)).pCorners
	end
end


local N_UD_EDGES_PERMUTATIONS = Permutation.number(8)

local pUDEdgesMove = {}
for i = 0, N_UD_EDGES_PERMUTATIONS - 1 do
	local state = indices_to_state({ pUDEdges = i })

	pUDEdgesMove[i] = {}
	for move_name, move in pairs(moves) do
		pUDEdgesMove[i][move_name] =
			state_to_indices(state:mul(move)).pUDEdges
	end
end


local N_E_EDGES_PERMUTATIONS = Permutation.number(4)

local pEEdgesMove = {}
for i = 0, N_E_EDGES_PERMUTATIONS - 1 do
	local state = indices_to_state({ pEEdges = i })

	pEEdgesMove[i] = {}
	for move_name, move in pairs(moves) do
		pEEdgesMove[i][move_name] =
			state_to_indices(state:mul(move)).pEEdges
	end
end


-- distance tables

local goal = state_to_indices(State.id)

local pCornersDistance = {}
for i = 0, N_CORNERS_PERMUTATIONS - 1 do
	pCornersDistance[i] = {}
end

pCornersDistance[goal.pCorners][goal.pEEdges] = 0

local distance = 0
repeat
	local n_visited = 0

	for i = 0, N_CORNERS_PERMUTATIONS - 1 do
		for j = 0, N_E_EDGES_PERMUTATIONS - 1 do
			if pCornersDistance[i][j] == distance then
				for move_name, _ in pairs(moves) do
					local next_p_corners = pCornersMove[i][move_name]
					local next_p_e_edges = pEEdgesMove[j][move_name]
					if not pCornersDistance[next_p_corners][next_p_e_edges] then
						pCornersDistance[next_p_corners][next_p_e_edges] = distance + 1
						n_visited = n_visited + 1
					end
				end
			end
		end
	end

	distance = distance + 1
until n_visited == 0


local pUDEdgesDistance = {}
for i = 0, N_UD_EDGES_PERMUTATIONS - 1 do
	pUDEdgesDistance[i] = {}
end

pUDEdgesDistance[goal.pUDEdges][goal.pEEdges] = 0

local distance = 0
repeat
	local n_visited = 0

	for i = 0, N_UD_EDGES_PERMUTATIONS - 1 do
		for j = 0, N_E_EDGES_PERMUTATIONS - 1 do
			if pUDEdgesDistance[i][j] == distance then
				for move_name, _ in pairs(moves) do
					local next_p_edges = pUDEdgesMove[i][move_name]
					local next_p_e_edges = pEEdgesMove[j][move_name]
					if not pUDEdgesDistance[next_p_edges][next_p_e_edges] then
						pUDEdgesDistance[next_p_edges][next_p_e_edges] = distance + 1
						n_visited = n_visited + 1
					end
				end
			end
		end
	end

	distance = distance + 1
until n_visited == 0


-- search

Phase2 = IDAStar:new()

function Phase2:solved(state)
	return state.pCorners == goal.pCorners and
           state.pUDEdges == goal.pUDEdges and
		   state.pEEdges == goal.pEEdges
end

function Phase2:prune(state, depth)
	return pCornersDistance[state.pCorners][state.pEEdges] > depth or
	       pUDEdgesDistance[state.pUDEdges][state.pEEdges] > depth
end

local axis = {
	["U"] = "U", ["U2"] = "U", ["U'"] = "U",
	["D"] = "D", ["D2"] = "D", ["D'"] = "D",
	["L2"] = "L",
	["R2"] = "R",
	["F2"] = "F",
	["B2"] = "B",
}

function Phase2:next_states(state, path)
	return coroutine.wrap(function()
		local last = path[table.maxn(path)]

		for move_name, move in pairs(moves) do
			if not last or axis[last] ~= axis[move_name] then
				local next_state = {
					pCorners = pCornersMove[state.pCorners][move_name],
					pUDEdges = pUDEdgesMove[state.pUDEdges][move_name],
					pEEdges = pEEdgesMove[state.pEEdges][move_name],
				}

				coroutine.yield(move_name, next_state)
			end
		end
	end)
end

function Phase2:preprocess_state(state)
	return state_to_indices(state)
end
