package com.puzzletimer.solvers;

public class Solver {
    // constants
    private static int N_CORNERS_PERMUTATIONS = 40320;
    private static int N_CORNERS_ORIENTATIONS = 2187;
    private static int N_EDGES_COMBINATIONS = 924;
    private static int N_EDGES_PERMUTATIONS = 720;
    private static int N_EDGES_ORIENTATIONS = 2048;

    // indexing
    public static int[] stateToIndices(State s) {
        // edges 1
        boolean[] edges1 = {
            true,  true,  true,  true,
            true,  true,  false, false,
            false, false, false, false,
        };

        byte[] edges1Mapping = {
            0,  1,  2,  3,
            4,  5, -1, -1,
           -1, -1, -1, -1,
        };

        byte[] edges2Mapping = {
            -1, -1, -1, -1,
            -1, -1,  0,  1,
             2,  3,  4,  5,
        };

        boolean[] cEdges1 = new boolean[12];
        for (int i = 0; i < cEdges1.length; i++) {
            cEdges1[i] = edges1[s.pEdges[i]];
        }

        int[] pEdges1 = new int[6];
        int next1 = 0;
        int[] pEdges2 = new int[6];
        int next2 = 0;
        for (int i = 0; i < s.pEdges.length; i++) {
            if (cEdges1[i]) {
                pEdges1[next1] = edges1Mapping[s.pEdges[i]];
                next1++;
            } else {
                pEdges2[next2] = edges2Mapping[s.pEdges[i]];
                next2++;
            }
        }

        return new int[] {
            Permutation.toIndex(s.pCorners),
            ZeroSumOrientation.toIndex(s.oCorners, 3),
            Combination.toIndex(cEdges1, 6),
            Permutation.toIndex(pEdges1),
            Permutation.toIndex(pEdges2),
            Orientation.toIndex(s.oCorners, 2),
        };
    }

    public static State indicesToState(int[] indices) {
        boolean[] cEdges1 = Combination.fromIndex(indices[2], 6, 12);
        int[] pEdges1 = Permutation.fromIndex(indices[3], 6);
        int[] pEdges2 = Permutation.fromIndex(indices[4], 6);

        int[] edges1 = { 0, 1, 2, 3, 4, 5 };
        int next1 = 0;
        int[] edges2 = { 6, 7, 8, 9, 10, 11 };
        int next2 = 0;

        int[] pEdges = new int[12];
        for (int i = 0; i < pEdges.length; i++) {
            if (cEdges1[i]) {
                pEdges[i] = edges1[pEdges1[next1]];
                next1++;
            } else {
                pEdges[i] = edges2[pEdges2[next2]];
                next2++;
            }
        }

        return new State(
            Permutation.fromIndex(indices[0], 8),
            ZeroSumOrientation.fromIndex(indices[1], 3, 8),
            pEdges,
            ZeroSumOrientation.fromIndex(indices[5], 2, 12));
    }

    // moves
    private static State[] moves;
    private static String[] moveNames;

    static {
        moveNames = new String[] {
            "U", "U2", "U'",
            "D", "D2", "D'",
            "L", "L2", "L'",
            "R", "R2", "R'",
            "F", "F2", "F'",
            "B", "B2", "B'",
        };

        moves = new State[moveNames.length];
        for (int i = 0; i < moves.length; i++) {
            moves[i] = State.moves.get(moveNames[i]);
        }
    }

    // move tables
    private static int[][] pCornersMove;
    private static int[][] oCornersMove;
    private static int[][] pEdges1Move;
    private static int[][] pEdges2Move;
    private static int[][] oEdgesMove;

    static {
        pCornersMove = new int[N_CORNERS_PERMUTATIONS][moves.length];
        for (int i = 0; i < pCornersMove.length; i++) {
            State s = indicesToState(new int[] { i, 0, 0, 0, 0, 0 });

            for (int j = 0; j < moves.length; j++) {
                pCornersMove[i][j] = stateToIndices(s.mul(moves[j]))[0];
            }
        }

        oCornersMove = new int[N_CORNERS_ORIENTATIONS][moves.length];
        for (int i = 0; i < oCornersMove.length; i++) {
            State s = indicesToState(new int[] { 0, i, 0, 0, 0, 0 });

            for (int j = 0; j < moves.length; j++) {
                oCornersMove[i][j] = stateToIndices(s.mul(moves[j]))[1];
            }
        }

        pEdges1Move = new int[N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS][moves.length];
        pEdges2Move = new int[N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS][moves.length];
        for (int i = 0; i < N_EDGES_COMBINATIONS; i++) {
            for (int j = 0; j < N_EDGES_PERMUTATIONS; j++) {
                State state = indicesToState(new int[] { 0, 0, i, j, j, 0 });
                for (int k = 0; k < moves.length; k++) {
                    int[] indices = stateToIndices(state.mul(moves[k]));
                    pEdges1Move[i * N_EDGES_PERMUTATIONS + j][k] =
                        indices[2] * N_EDGES_PERMUTATIONS + indices[3];
                    pEdges2Move[i * N_EDGES_PERMUTATIONS + j][k] =
                        indices[2] * N_EDGES_PERMUTATIONS + indices[4];
                }
            }
        }

        oEdgesMove = new int[N_EDGES_ORIENTATIONS][moves.length];
        for (int i = 0; i < N_EDGES_ORIENTATIONS; i++) {
            State state = indicesToState(new int[] { 0, 0, 0, 0, 0, i });

            for (int j = 0; j < moves.length; j++) {
                oEdgesMove[i][j] = stateToIndices(state.mul(moves[j]))[5];
            }
        }
    }

    // goal
    private static int[] goal;

    static {
        goal = stateToIndices(State.id);
    }

    // distance tables
    private static byte[][] cornersDistance;
    private static byte[] edges1Distance;
    private static byte[] edges2Distance;

    static {
        cornersDistance = new byte[N_CORNERS_PERMUTATIONS][N_CORNERS_ORIENTATIONS];
        for (int i = 0; i < cornersDistance.length; i++) {
            for (int j = 0; j < cornersDistance[i].length; j++) {
                cornersDistance[i][j] = -1;
            }
        }
        cornersDistance[goal[0]][goal[1]] = 0;

        int distance = 0;
        int nVisited = 1;
        while (nVisited < N_CORNERS_PERMUTATIONS * N_CORNERS_ORIENTATIONS) {
            int n = 0;
            for (int i = 0; i < N_CORNERS_PERMUTATIONS; i++) {
                for (int j = 0; j < N_CORNERS_ORIENTATIONS; j++) {
                    if (cornersDistance[i][j] == distance) {
                        for (int k = 0; k < moves.length; k++) {
                            int pCornersNext = pCornersMove[i][k];
                            int oCornersNext = oCornersMove[j][k];
                            if (cornersDistance[pCornersNext][oCornersNext] < 0) {
                                cornersDistance[pCornersNext][oCornersNext] = (byte) (distance + 1);
                                nVisited++;
                                n++;
                            }
                        }
                    }
                }
            }

            distance++;
        }

        edges1Distance = new byte[N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS];
        for (int i = 0; i < edges1Distance.length; i++) {
            edges1Distance[i] = -1;
        }
        edges1Distance[goal[2] * N_EDGES_PERMUTATIONS + goal[3]] = 0;

        distance = 0;
        nVisited = 1;
        while (nVisited < N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS) {
            for (int i = 0; i < N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS; i++) {
                if (edges1Distance[i] == distance) {
                    for (int k = 0; k < moves.length; k++) {
                        int pEdges1Next = pEdges1Move[i][k];
                        if (edges1Distance[pEdges1Next] < 0) {
                            edges1Distance[pEdges1Next] = (byte) (distance + 1);
                            nVisited++;
                        }
                    }
                }
            }
            distance++;
        }

        edges2Distance = new byte[N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS];
        for (int i = 0; i < edges2Distance.length; i++) {
            edges2Distance[i] = -1;
        }
        edges2Distance[goal[2] * N_EDGES_PERMUTATIONS + goal[4]] = 0;

        distance = 0;
        nVisited = 1;
        while (nVisited < N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS) {
            for (int i = 0; i < N_EDGES_COMBINATIONS * N_EDGES_PERMUTATIONS; i++) {
                if (edges2Distance[i] == distance) {
                    for (int k = 0; k < moves.length; k++) {
                        int pEdges2Next = pEdges2Move[i][k];
                        if (edges2Distance[pEdges2Next] < 0) {
                            edges2Distance[pEdges2Next] = (byte) (distance + 1);
                            nVisited++;
                        }
                    }
                }
            }
            distance++;
        }
    }

    // moves
    private static int[] sides;
    private static int[] axes;

    static {
        sides = new int[] {
            0, 0, 0,
            1, 1, 1,
            2, 2, 2,
            3, 3, 3,
            4, 4, 4,
            5, 5, 5,
        };

        axes = new int[] {
            0, 0, 0,
            0, 0, 0,
            1, 1, 1,
            1, 1, 1,
            2, 2, 2,
            2, 2, 2,
        };
    }

    // search
    public static String[] solution(State state) {
        int[] indices = stateToIndices(state);

        for (int depth = 0; ; depth++) {
            int[] path = new int[depth];
            if (search(indices[0],
                       indices[1],
                       indices[2] * N_EDGES_PERMUTATIONS + indices[3],
                       indices[2] * N_EDGES_PERMUTATIONS + indices[4],
                       indices[5],
                       depth,
                       path)) {
                String[] sequence = new String[path.length];
                for (int i = 0; i < sequence.length; i++) {
                    sequence[i] = moveNames[path[i]];
                }

                return sequence;
            }
        }
    }

    private static boolean search(int pCorners, int oCorners, int pEdges1, int pEdges2, int oEdges, int depth, int[] path) {
        if (depth == 0) {
            return pCorners == goal[0] &&
                   oCorners == goal[1] &&
                   pEdges1 == goal[2] * N_EDGES_PERMUTATIONS + goal[3] &&
                   pEdges2 == goal[2] * N_EDGES_PERMUTATIONS + goal[4] &&
                   oEdges == goal[5];
        }

        if (cornersDistance[pCorners][oCorners] > depth ||
            edges1Distance[pEdges1] > depth ||
            edges2Distance[pEdges2] > depth) {
            return false;
        }

        int lastSide = -1;
        if (path.length - depth - 1 >= 0) {
            lastSide = sides[path[path.length - depth - 1]];
        }

        int[] lastAxes = { -1, -1 };
        for (int i = 0; i < lastAxes.length; i++) {
            if (path.length - depth - 1 - i >= 0) {
                lastAxes[i] = axes[path[path.length - depth - 1 - i]];
            }
        }

        for (int i = 0; i < moves.length; i++) {
            if (sides[i] == lastSide ||
                (axes[i] == lastAxes[0] && axes[i] == lastAxes[1])) {
                continue;
            }

            path[path.length - depth] = i;
            if (search(pCornersMove[pCorners][i],
                       oCornersMove[oCorners][i],
                       pEdges1Move[pEdges1][i],
                       pEdges2Move[pEdges2][i],
                       oEdgesMove[oEdges][i],
                       depth - 1,
                       path)) {
                return true;
            }
        }

        return false;
    }
}
